/*
 * imap_err.h:
 * This file is automatically generated; please do not edit it.
 */

#include <et/com_err.h>

#define IMAP_INTERNAL                            (-1904809472L)
#define IMAP_IOERROR                             (-1904809471L)
#define IMAP_NOTFOUND                            (-1904809470L)
#define IMAP_SYS_ERROR                           (-1904809469L)
#define IMAP_NOSPACE                             (-1904809468L)
#define IMAP_PERMISSION_DENIED                   (-1904809467L)
#define IMAP_QUOTA_EXCEEDED                      (-1904809466L)
#define IMAP_MESSAGE_TOO_LARGE                   (-1904809465L)
#define IMAP_USERFLAG_EXHAUSTED                  (-1904809464L)
#define IMAP_NAMESPACE_BADPREFIX                 (-1904809463L)
#define IMAP_MAILBOX_BADFORMAT                   (-1904809462L)
#define IMAP_SYNC_CHECKSUM                       (-1904809461L)
#define IMAP_MAILBOX_CHECKSUM                    (-1904809460L)
#define IMAP_MAILBOX_NOTSUPPORTED                (-1904809459L)
#define IMAP_MAILBOX_NONEXISTENT                 (-1904809458L)
#define IMAP_MAILBOX_EXISTS                      (-1904809457L)
#define IMAP_MAILBOX_BADNAME                     (-1904809456L)
#define IMAP_MAILBOX_BADTYPE                     (-1904809455L)
#define IMAP_MAILBOX_MOVED                       (-1904809454L)
#define IMAP_MAILBOX_RESERVED                    (-1904809453L)
#define IMAP_MAILBOX_LOCKED                      (-1904809452L)
#define IMAP_MAILBOX_DISABLED                    (-1904809451L)
#define IMAP_PARTITION_UNKNOWN                   (-1904809450L)
#define IMAP_INVALID_IDENTIFIER                  (-1904809449L)
#define IMAP_MESSAGE_CONTAINSNULL                (-1904809448L)
#define IMAP_MESSAGE_CONTAINSNL                  (-1904809447L)
#define IMAP_MESSAGE_CONTAINS8BIT                (-1904809446L)
#define IMAP_MESSAGE_BADHEADER                   (-1904809445L)
#define IMAP_MESSAGE_NOBLANKLINE                 (-1904809444L)
#define IMAP_QUOTAROOT_NONEXISTENT               (-1904809443L)
#define IMAP_PROTOCOL_ERROR                      (-1904809442L)
#define IMAP_PROTOCOL_BAD_PARAMETERS             (-1904809441L)
#define IMAP_ANNOTATION_BADENTRY                 (-1904809440L)
#define IMAP_ANNOTATION_BADATTRIB                (-1904809439L)
#define IMAP_ANNOTATION_BADVALUE                 (-1904809438L)
#define IMAP_BADURL                              (-1904809437L)
#define IMAP_ZERO_LENGTH_LITERAL                 (-1904809436L)
#define IMAP_BAD_SERVER                          (-1904809435L)
#define IMAP_SERVER_UNAVAILABLE                  (-1904809434L)
#define IMAP_REMOTE_DENIED                       (-1904809433L)
#define IMAP_AGAIN                               (-1904809432L)
#define IMAP_NOT_SINGULAR_ROOT                   (-1904809431L)
#define IMAP_REMOTE_NO_MULTIAPPEND               (-1904809430L)
#define IMAP_UNRECOGNIZED_CHARSET                (-1904809429L)
#define IMAP_INVALID_USER                        (-1904809428L)
#define IMAP_INVALID_LOGIN                       (-1904809427L)
#define IMAP_ANONYMOUS_NOT_PERMITTED             (-1904809426L)
#define IMAP_UNSUPPORTED_QUOTA                   (-1904809425L)
#define IMAP_SASL_FAIL                           (-1904809424L)
#define IMAP_SASL_CANCEL                         (-1904809423L)
#define IMAP_SASL_PROTERR                        (-1904809422L)
#define IMAP_NO_OVERQUOTA                        (-1904809421L)
#define IMAP_NO_CLOSEQUOTA                       (-1904809420L)
#define IMAP_NO_MSGGONE                          (-1904809419L)
#define IMAP_NO_CHECKSEEN                        (-1904809418L)
#define IMAP_NO_CHECKPRESERVE                    (-1904809417L)
#define IMAP_NO_NOSUCHMSG                        (-1904809416L)
#define IMAP_NO_NOSUCHANNOTATION                 (-1904809415L)
#define IMAP_NO_UNKNOWN_CTE                      (-1904809414L)
#define IMAP_BYE_LOGOUT                          (-1904809413L)
#define IMAP_OK_COMPLETED                        (-1904809412L)
extern const struct error_table et_imap_error_table;
extern void initialize_imap_error_table(void);

/* For compatibility with Heimdal */
extern void initialize_imap_error_table_r(struct et_list **list);

#define ERROR_TABLE_BASE_imap (-1904809472L)

/* for compatibility with older versions... */
#define init_imap_err_tbl initialize_imap_error_table
#define imap_err_base ERROR_TABLE_BASE_imap
